//
//  tfp.c
//  tfp - plays floating point data streams in text format
//
//  Created by Stanko Juzbasic on 11/25/13.
//  Copyright (c) 2013. All rights reserved.
//

/*  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  
 *          T   E   X   T          F   I   L   E          P   L   A   Y   E   R
 *          Prototype code for sonifying arbitrary Float32 array data on MacOSX
 *          -   -   -   -   -  - - directly using AudioUnit API- -  -   -   -   -   -   - 
 *          -   -   -   -   -   -   - 2013, Stanko Juzbasic -   -   -   -   -   -   -   -  
 *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  */
//________________________________________________________________________________________
#pragma mark - headers -
//________________________________________________________________________________________


#include <AudioToolbox/AudioToolbox.h>
#include <unistd.h>
//________________________________________________________________________________________
#pragma mark - globals & prototypes -
//________________________________________________________________________________________



Float64 R; 
Float32 N;

typedef struct FloatBufPlayer
{
	AudioUnit outputUnit;
	double startingFrameCount;
    Float32* fBuffer;
} FloatBufPlayer;

OSStatus FloatBufRenderProc(void *inRefCon,
							AudioUnitRenderActionFlags *ioActionFlags,
							const AudioTimeStamp *inTimeStamp,
							UInt32 inBusNumber,
							UInt32 inNumberFrames,
							AudioBufferList * ioData);

void CreateAndConnectFpOutputUnit (FloatBufPlayer *fplayer) ;

size_t  getSize(FILE*, char*);
void    normalize(Float32* , size_t);
Float32 kramp ( Float32 );
int     kbhit(void);
void    usage (void);

double  totalFrames;
Boolean PSTOP = FALSE;
Float32* fBuffer;
size_t  fbuffersize;

Float32 N;
Boolean STOP = FALSE;

//________________________________________________________________________________________
#pragma mark - auxilliary functions -
//________________________________________________________________________________________
size_t getSize(FILE* file, char* name)
{   
    size_t s;
    if(file == NULL)return 0;
    fseek(file, 0L, SEEK_END);
    s = ftell(file);   
    s = s/sizeof(Float32);
    printf("File %s: %zu\t values\n\n",name, s);
    rewind(file);
    return s;
}
//________________________________________________________________________________________
// this is a UNIX replacement for non-blocking keyboard hit detector
//________________________________________________________________________________________
int kbhit(void)
{
    
    struct timeval tv;    
    fd_set read_fd;
    
    tv.tv_sec=0;    
    tv.tv_usec=0;
    
    FD_ZERO(&read_fd);
    
    FD_SET(0,&read_fd);
    
    if(select(1, &read_fd, NULL, NULL, &tv) == -1)        
        return 0;
    
    if(FD_ISSET(0,&read_fd))        
        return 1;
    
    return 0;
    
}
//________________________________________________________________________________________
//________________________________________________________________________________________
void normalize(Float32* buf, size_t siz)
{   
    Float32 m;
    for (int i = 0; i < siz; i++) 
        if (fabsf(buf[i]) > m ) {
            m = fabsf(buf[i]);
        }
    m = .9/m;
    for (int j = 0 ; j < siz; j++) 
        buf[j] = m*buf[j];
    
    return;  
}
//________________________________________________________________________________________
//________________________________________________________________________________________
//generates 1000 ramp samples (it's a hack!)
Float32 kramp ( Float32 i)
{
    if(i >= 2000.)return 1.;
    if(i <= 0.)return 0.;
    return .0005*(i);
}
//________________________________________________________________________________________
#pragma mark - callback function -
//________________________________________________________________________________________
OSStatus FloatBufRenderProc(void *inRefCon,
							AudioUnitRenderActionFlags *ioActionFlags,
							const AudioTimeStamp *inTimeStamp,
							UInt32 inBusNumber,
							UInt32 inNumberFrames,
							AudioBufferList * ioData)
{

    if(fBuffer == NULL){
        printf("Can't access input buffer\n");
        return 0;
    }
    
	FloatBufPlayer *fplayer = (FloatBufPlayer*)inRefCon;
	
	double j = fplayer->startingFrameCount;
    long   l = (long)j;
 

    Float32 *outL = (Float32*)ioData->mBuffers[0].mData;
    Float32 *outR = (Float32*)ioData->mBuffers[1].mData;       
    
	int  frame = 0;
	for (frame = 0; frame < inNumberFrames; ++frame) 
	{            
        
        //copy data to buffer
        if(STOP){
        (outL)[frame] =  fBuffer[l]*kramp(N--);    
        (outR)[frame] =  fBuffer[l]*kramp(N--);            
        }else if(totalFrames >= fbuffersize){
        (outL)[frame] =  0.;    
        (outR)[frame] =  0.;        
        }else{
        (outL)[frame] =  fBuffer[l];    
        (outR)[frame] =  fBuffer[l];
        }
        
		j += 1.0;
        totalFrames +=1.;
        l++;
	}//end for(frame = 0;...
	
	fplayer->startingFrameCount = j;
	return noErr;
}


//________________________________________________________________________________________
#pragma mark - utility functions -
//________________________________________________________________________________________
// generic error handler - if err is nonzero, prints error message and exits program.
static void CheckError(OSStatus error, const char *operation)
{
	if (error == noErr) return;
	
	char str[20];
	// see if it appears to be a 4-char-code
	*(UInt32 *)(str + 1) = CFSwapInt32HostToBig(error);
	if (isprint(str[1]) && isprint(str[2]) && isprint(str[3]) && isprint(str[4])) {
		str[0] = str[5] = '\'';
		str[6] = '\0';
	} else
		// no, format it as an integer
		sprintf(str, "%d", (int)error);
	
    //	fprintf(stderr, "Error: %s (%s)\n", operation, str);
    printf( "Error: %s (%s)\n", operation, str);
	
	exit(1);
}
//________________________________________________________________________________________
//________________________________________________________________________________________

void CreateAndConnectFpOutputUnit (FloatBufPlayer *fplayer) 
{
	UInt32 size = sizeof(Float64);
	//  10.6 and later: generate description that will match out default output device 
	AudioComponentDescription theoutput = {0}; // 10.6 version
	theoutput.componentType = kAudioUnitType_Output;
	theoutput.componentSubType = kAudioUnitSubType_DefaultOutput;
	theoutput.componentManufacturer = kAudioUnitManufacturer_Apple;
	
	AudioComponent comp = AudioComponentFindNext (NULL, &theoutput);
	if (comp == NULL) {
		printf ("can't get output unit");
		exit (-1);
	}
	CheckError (AudioComponentInstanceNew(comp, &fplayer->outputUnit),
				"Couldn't open component for outputUnit");    
    
    
	// register render callback
	AURenderCallbackStruct input;   
	input.inputProc = FloatBufRenderProc; //name of the callback process
	input.inputProcRefCon = fplayer;      //name of the unit
    
    // set render callback
	CheckError(AudioUnitSetProperty(fplayer->outputUnit,
									kAudioUnitProperty_SetRenderCallback, 
									kAudioUnitScope_Input,
									0,
									&input, 
									sizeof(input)),
			   "AudioUnitSetProperty failed");
    
    // get sample rate
    CheckError(AudioUnitGetProperty(fplayer->outputUnit,                         
                                    kAudioUnitProperty_SampleRate,
                                    kAudioUnitScope_Output,
                                    0,
                                    &R, 
                                    &size),
               "AudioUnitGetProperty failed");
    
	// initialize unit
	CheckError (AudioUnitInitialize(fplayer->outputUnit),
				"Couldn't initialize output unit");
    
}
//________________________________________________________________________________________
#pragma mark - usage -
//________________________________________________________________________________________
void usage(void)
{
    printf("\nUSAGE:\n");
    printf("tfp <filename>.txt\n");
    printf("\n(file has to be text format w. float values)\n\n");
    return;
}

//________________________________________________________________________________________
#pragma mark - main -
//________________________________________________________________________________________


int main (int argc, const char * argv[])
{

    N = 2000.;
    FILE *afPtr;
    char afname[512];
    char text[16];
    size_t size;
    
    printf("\n______________________________________________\n");
    printf("* * * * Text File Player for CoreAudio * * * *\n");
    if (argv[1] == NULL)
        goto no_name;
    
    strcpy(afname, argv[1]);
    printf("\nName:%s\tFile: %s\n", argv[1], afname);
    afPtr = fopen(afname,"rb");

    
    if (afPtr == NULL){
        printf("File \"%s\"\tnot found\n", argv[1]);
    no_name:
        usage();
        exit(0);
    }
    
    size = 1;
    while( !feof(afPtr)) {
        fscanf(afPtr, "%s",text);
        size++;
    }
    
    rewind(afPtr);
    fBuffer = calloc((size+1), sizeof(Float32));    
    fbuffersize = size;
    
    int i = 0;    
    while( !feof(afPtr)) {
        fscanf(afPtr, "%s",text);
        Float32 val = (Float32)atof(text);
        fBuffer[i++] = val;
    }

    normalize(fBuffer, fbuffersize);
    
    totalFrames = 0.;
    N = 2000.;
    fclose(afPtr);
     
 	FloatBufPlayer fplayer = {0};
    
    fplayer.fBuffer = fBuffer;
	
	// set up unit and callback
	CreateAndConnectFpOutputUnit(&fplayer);

	// start playing
	CheckError (AudioOutputUnitStart(fplayer.outputUnit), "Couldn't start output unit");
    
    printf ("\nPress \"Enter\" to stop:\n");
	
    //play until done or stopped
    while(totalFrames < fbuffersize)
        if (kbhit()){
            STOP = TRUE;
            break;
            }
    
    //wait for the playback to die out
    if (STOP)usleep(200000);
 
    
cleanup:
	AudioOutputUnitStop(fplayer.outputUnit);
	AudioUnitUninitialize(fplayer.outputUnit);
	AudioComponentInstanceDispose(fplayer.outputUnit);
    printf("Frames Played: %8.0lf\n", totalFrames);
    
    printf("TFP: done!\n");
    printf("_____________________________________________\n\n");    
    free(fBuffer);
   
    return 0;
}

